#ifndef DLX_ALGORITHM_HPP
#define DLX_ALGORITHM_HPP



#include <limits>
#include <map>
#include <vector>




namespace soma
  {



  namespace solver
    {



    namespace dlx_algorithm
      {



      template
        <
        typename DerivedT
        >
      class CDLXAlgorithm
        {

        private :

//  Copie d'un objet de type CDLXAlgorithm interdite.
          CDLXAlgorithm
                      (
                      const CDLXAlgorithm&
                      );



//  Assignation d'un objet de type CAllPolyformPiecesPositions interdite.
          void operator=
                       (
                       const CDLXAlgorithm&
                       );


//  Type de noeuds servant de base aux types header_t et matrix_element_t.
//  Chaque noeud de type base_node_t possde un lien vers 4 noeuds du mme type
//  situs  sa gauche, sa droite, au-dessus et en-dessoous de lui, pour former,
//  in fine, une structure toroidale,  bz=ase de listes doublement chnes circulaires.
          struct TBaseNode
            {

            TBaseNode *down,
                      *left,
                      *right,
                      *up;

            };



//  Le type header_t drive de base_node_t et modlise les en-ttes de la structure toroidale.
//  Leur rle est de lister tous les noeuds d'une colonne.
          struct THeader : TBaseNode
            {

            char name;


            unsigned int column,
                         size;

            };



//  matrix_element_t est le type des noeuds intrieurs de la structure toroidale. Ils modlisent
//  en fait, les 1 de la matrice boolenne de dpart.
          struct TMatrixElement : TBaseNode
            {

            unsigned int row;

            THeader* header;

            };


        protected :

          typedef TBaseNode base_node_t;
          typedef THeader header_t;
          typedef TMatrixElement matrix_element_t;


        private :

          const unsigned int m_MaxNbSolutions;



/*
  Fonction membre prive cover. Voir le fichier d'implmentation DLXAlgorithm.tpp pour
  avoir une description plus complte de la fonction cover.

  Argument :
    - type : header_t* const.
    - nom : col.

  Retour : void.
*/
          inline void cover
                          (
                          header_t* const col
                          );



/*
  Fonction membre prive uncover. Voir le fichier d'implmentation DLXAlgorithm.tpp pour
  avoir une description plus complte de la fonction uncover.

  Argument :
    - type : header_t* const.
    - nom : col.

  Retour : void.
*/

          inline void uncover
                            (
                            header_t* const col
                            );



/*
  Fonction membre prive chooseColumn. Voir le fichier d'implmentation DLXAlgorithm.tpp pour
  avoir une description plus complte de la fonction chooseColumn.

  Argument : aucun.

  Retour :
    - type :header_t*.
    - nom : col.
    - col liste le moins de noeuds (donne membre size la plus petite)
*/

          inline header_t* chooseColumn();



/*  Fonction membre protge generateSparseMatrix. generateSparseMatrix est une fonction virtuelle
    (au sens du CRTP) devant tre redfinie dans une classe fille de CDLXAlgorithm. Son rle est de
    construire une structure toroidale  base de listes doublement chanes, en fonction des
    informations fournies par son argument template data (gnralement une matrice boolenne).

  Argument :
    - type : template<typename> DataT.
    - nom : data.
    - data contient les informations ncessaires pour pouvoir gnrer une matrice creuse.
*/
          template
            <
            typename DataT
            >
          inline void generateSparseMatrix
                                         (
                                         const DataT& data
                                         )
          {
          (static_cast<DerivedT*>(this))->generateSparseMatrix
                                                             (
                                                             data
                                                             );
          }


        protected :

// Racine de la structure toroidale :
          header_t* m_H;

// En-ttes de la structure toroidale :
          std::vector<header_t> m_Headers;

// Noeuds "intrieurs" de la structure toroidale.
          std::vector<matrix_element_t> m_MatrixElements;

// Pointeurs vers les noeuds de la structure toroidale runies par ligne comme
// dans la matrice boolenne originelle. Utile pour rcuprer une solution.
          std::vector<std::vector<matrix_element_t*> > m_Rows;

// Contient une solution de l'algorithme DLX :
          std::vector<unsigned int> m_O;

// Contient toutes les solutions de l'algorithme DLX :
          std::vector<std::vector<unsigned int> > m_Solutions;


        public :

          template
            <
            typename DataT
            >
          CDLXAlgorithm
                      (
                      const DataT& data,
                      const unsigned int maxNbSolutions=std::numeric_limits<unsigned int>::max()
                      );



/*
  Fonction membre publique search. Voir le fichier d'implmentation DLXAlgorithm.tpp pour
  avoir une description plus complte de la fonction search.

  Argument : aucun.

  Retour : void.
*/

          inline void search();



/*
  Fonction membre publique const hasSolution.

  Argument : Aucun.

  Retour
    - type : const bool.
    - Retourne true si l'algorithme DLX a trouv au moins 1 solution, faux sinon.
*/
          inline const bool hasSolution() const
            {
            return
                 (
                 m_Solutions.size()
                 );
            }



/*
  Fonction membre publique const getNumberOfFoundSolutions.

  Argument : Aucun.

  Retour
    - type : const unsigned int.
    - Retourne le nombre de solutions trouves (ou pas) par l'algorithme DLX.
*/
          inline const unsigned int getNumberOfFoundSolutions() const
            {
            return
                 (
                 m_Solutions.size()
                 );
            }



/*
  Fonction membre publique getIthSolution. Voir le fichier d'implmentation DLXAlgorithm.tpp pour
  avoir une description plus complte de la fonction getIthSolution.

  Argument : aucun.

  Retour :
    - type : const std::map<char,std::vector<unsigned int> >.
    - nom : ithSolution.
*/

          inline const std::map<char,std::vector<unsigned int> > getIthSolution
                                                                              (
                                                                              const unsigned int solutionIndex
                                                                              ) const;

        };



      #include "DLXAlgorithm.tpp"



      }



    }



  }



#endif  //  DLX_ALGORITHM_HPP
